// stdafx.h : include file for standard system include files,
// or project specific include files that are used frequently, but
// are changed infrequently
//

#pragma once

#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <tchar.h>

#include <vector>
#include <algorithm>

#include <iostream>
#include <iomanip>

#include <fstream>

#include <comdef.h>

#undef max
#undef min

#pragma warning(disable : 28020 26495 26812 26437 28182 6011)
#include <pybind11/pybind11.h>


//#include "TTCG/Common/Path.h"
